#![cfg(feature = "server")]
use std::{
    ops::Add,
    sync::{atomic::AtomicUsize, Arc},
    time::Duration,
};

use libpt::log::{debug, info, trace, warn};
use tokio::{
    io::{AsyncBufReadExt, AsyncWriteExt, BufReader},
    net::{TcpListener, TcpStream},
    time::{self, timeout},
};

use crate::common::conf::Config;

pub mod errors;
use errors::*;

pub struct Server {
    cfg: Config,
    pub timeout: Option<Duration>,
    server: TcpListener,
    num_peers: AtomicUsize,
}

impl Server {
    pub async fn build(cfg: Config) -> anyhow::Result<Self> {
        let server = TcpListener::bind(cfg.addr).await?;
        let timeout = Some(Duration::from_secs(5));
        let num_peers = AtomicUsize::new(0);
        Ok(Server {
            cfg,
            timeout,
            server,
            num_peers,
        })
    }
    pub async fn run(self) -> anyhow::Result<()> {
        let rc_self = Arc::new(self);
        let ref_self = rc_self.clone();
        tokio::spawn(async move {
            let mut interval = time::interval(Duration::from_millis(5000));
            loop {
                interval.tick().await;
                info!(
                    "status: {} peers",
                    ref_self
                        .num_peers
                        .load(std::sync::atomic::Ordering::Relaxed)
                );
            }
        });
        loop {
            let ref_self = rc_self.clone();
                let (stream, addr) = match ref_self.server.accept().await {
                    Ok(s) => s,
                    Err(err) => {
                        warn!("could not accept stream: {err:?}");
                        continue;
                    }
                };
            // NOTE: we can only start the task now. If we start it before accepting connections
            // (so that the task theoretically accepts the connection), we would create endless
            // tasks in a loop
            tokio::spawn(async move {
                ref_self.peer_add(1);
                match ref_self.handle_stream(stream).await {
                    Ok(_) => (),
                    Err(err) => match err {
                        ServerError::Timeout(_) => {
                            debug!("stream {:?} timed out", addr)
                        }
                        _ => {
                            warn!("error while handling stream: {:?}", err)
                        }
                    },
                };
                ref_self.peer_sub(1);
            });
        }
    }

    #[inline]
    fn peer_add(&self, v: usize) {
        self.num_peers.store(
            (self.num_peers.load(std::sync::atomic::Ordering::Relaxed) + v),
            std::sync::atomic::Ordering::Relaxed,
        )
    }

    #[inline]
    fn peer_sub(&self, v: usize) {
        self.num_peers.store(
            (self.num_peers.load(std::sync::atomic::Ordering::Relaxed) - v),
            std::sync::atomic::Ordering::Relaxed,
        )
    }

    async fn handle_stream(&self, stream: TcpStream) -> Result<()> {
        let mut pings: usize = 0;
        let addr = match stream.peer_addr() {
            Ok(a) => a,
            Err(err) => {
                debug!("could not get peer address: {:?}", err);
                return Err(err.into());
            }
        };
        debug!("new peer: {:?}", addr);
        let mut buf = Vec::new();
        let mut reader = BufReader::new(stream);
        loop {
            match self.read(&mut reader, &mut buf).await {
                Ok(len) if len == 0 => {
                    trace!("len is 0, so the stream has ended: {len:?}");
                    break;
                }
                Ok(len) => len,
                Err(err) => {
                    match err {
                        ServerError::Timeout(_) => {
                            debug!("peer {:?} timed out", addr)
                        }
                        _ => return Err(err),
                    }
                    break;
                }
            };
            trace!("received message: {:X?}", buf);
            let msg = self.decode(&buf)?;
            debug!("< {:?} : {}", addr, msg);
            if msg.contains("ping") {
                pings += 1;
            }
            if pings < self.cfg.win_after {
                reader.write_all(b"pong\0").await?;
                debug!("> {:?} : pong", addr,);
            } else {
                reader.write_all(b"you win!\0").await?;
                debug!("> {:?} : you win!", addr,);
                reader.shutdown().await?;
                break;
            }
            buf.clear();

            // we should wait, so that we don't spam the client
            std::thread::sleep(self.cfg.delay);
        }
        debug!("disconnected peer: {:?}", addr);
        Ok(())
    }

    #[inline]
    fn decode(&self, buf: &Vec<u8>) -> Result<String> {
        Ok(String::from_utf8(buf.clone())?.replace('\n', "\\n"))
    }

    #[inline]
    async fn read(&self, reader: &mut BufReader<TcpStream>, buf: &mut Vec<u8>) -> Result<usize> {
        match timeout(self.cfg.timeout, reader.read_until(0x00, buf)).await? {
            Ok(len) => Ok(len),
            Err(err) => Err(err.into()),
        }
    }
}
